//+------------------------------------------------------------------+
//|                                                   Dictionary.mqh |
//|                                                   Enrico Lambino |
//|                             https://www.mql5.com/en/users/iceron |
//| Ver 1.1:added serial iteration                                   |
//|         memory management                                        |
//+------------------------------------------------------------------+
#property copyright "Enrico Lambino."
#property link      "https://www.mql5.com/en/users/iceron"
#property strict
#include <Arrays\ArrayObj.mqh>
#include <Arrays\List.mqh>
//+------------------------------------------------------------------+
//| Dictionary for a single data type, key can be any primitive type.+
//| Duplicate key additions override the previous key.               +
//| Serial iteration in the order the objects are introduced to      +
//| dictionary is supported.                                         +
//| Example   CDictionary<double> dict;                              +
//|   for(int i=0;i<Bars;i++)                                        +
//|      dict.Add(Time[i],Close[i]);                                 +
//|   for(int i=0;i<Bars;i++)                                        +
//|      if(dict[Time[i]] != Close[i])                               +
//|         Print("Collision Error ",++cnt);                         +
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
class CDictionaryEntry : public CObject
  {
protected:
   T                 m_value;
   string            m_key;
public:
                     CDictionaryEntry(string key,T value);
                    ~CDictionaryEntry(void);
   //--- for serial search in order of addition
   CDictionaryEntry *m_next_entry;
   CDictionaryEntry *m_prev_entry;
   //--- methods of access to protected data
   void              Key(const string key);
   string            Key(void) const;
   bool              Value(string key,T value);
   T                 Value();
   virtual int       Compare(const CObject *node,const int mode=0) const override;
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
CDictionaryEntry::CDictionaryEntry(string key,T value)
  {
   m_value=value;
   m_key=key;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
CDictionaryEntry::~CDictionaryEntry(void)
  {
   CObject *object=dynamic_cast<CObject*>(m_value);
   if(CheckPointer(object)==POINTER_DYNAMIC)
      delete object;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
CDictionaryEntry::Key(const string key)
  {
   m_key=key;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
string CDictionaryEntry::Key(void) const
  {
   return m_key;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
int CDictionaryEntry::Compare(const CObject *node,const int mode=0) const
  {
   const CDictionaryEntry *node_entry=node;
   string str1 = Key();
   string str2 = node_entry.Key();
   return StringCompare(str1,str2);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
bool CDictionaryEntry::Value(string key,T value)
  {
   CObject *m_object=dynamic_cast<CObject*>(m_value);
   if(CheckPointer(m_object)==POINTER_DYNAMIC)
     {
      CObject *object=dynamic_cast<CObject*>(value);
      if(CheckPointer(object) && m_object!=object)
        {
         delete m_object;
         m_object=object;
        }
     }
   else m_value=value;
   return true;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
T CDictionaryEntry::Value()
  {
   return m_value;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
class CDictionary : public CObject
  {
protected:
   CList            *m_data[];           // data array
   uint              m_data_total;       // number of elements in dictionary (including in CList)
   uint              m_data_max;         // maximmum size of the array without memory reallocation
   uint              m_initial_capacity;
   double            m_resize_threshold; // threshold to resize - m_data_total>m_data_max*m_resize_threshold
   uint              m_multiplier;
   bool              m_free_mode;        // flag of necessity of "physical" deletion of object
   //--- locators for serial iteration
   CDictionaryEntry<T>*m_current_entry;
   CDictionaryEntry<T>*m_first_entry;
   CDictionaryEntry<T>*m_last_entry;
public:
                     CDictionary(uint initial_capacity=500,double resize_threshold=0.6,uint multiplier=3);
                    ~CDictionary(void);
   //--- methods of access to protected data
   bool              FreeMode(void) const { return(m_free_mode); }
   void              FreeMode(const bool mode);
   //--- methods of managing dynamic memory
   bool              Shutdown(void);
   bool              Add(CDictionaryEntry<T>*entry);
   template<typename KEY>
   bool              Add(KEY key,T value);
   template<typename KEY>
   bool              Contains(KEY key);
   template<typename KEY>
   bool              Delete(KEY key);
   template<typename KEY>
   T                 Get(KEY key);
   template<typename KEY>
   bool              Get(KEY key,T &value);
   bool              Reset();
   //template<typename KEY>
   //bool              Set(KEY key,T value);

   template<typename KEY>
   T  operator[](KEY key){ return Get(key); }

   void              Clear(void);
   //--- methods for serial iteration 
   int               Total(){ return m_data_total;}
   CDictionaryEntry<T>*GetNextNode(void);
   CDictionaryEntry<T>*GetPrevNode(void);
   CDictionaryEntry<T>*GetCurrentNode(void);
   CDictionaryEntry<T>*GetFirstNode(void);
   CDictionaryEntry<T>*GetLastNode(void);

protected:
   bool              Allocate(void);
   uint              Hash(string str);
   int               Index(uint key);
   bool              Recalculate();
   //--- methods for serial iteration
   bool              ReconnectChain(CDictionaryEntry<T>*deleted_item);
   bool              AddModelToChain(CDictionaryEntry<T>*model);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
CDictionary::CDictionary
(uint initial_capacity=500,double resize_threshold=0.6,uint multiplier=3) : m_initial_capacity(initial_capacity),
                                                                            m_resize_threshold(resize_threshold),
                                                                            m_multiplier(multiplier),
                                                                            m_current_entry(NULL),
                                                                            m_first_entry(NULL),
                                                                            m_last_entry(NULL),
                                                                            m_data_total(0),
                                                                            m_data_max(0),
                                                                            m_free_mode(true)
  {
   Reset();
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
CDictionary::~CDictionary(void)
  {
   if(m_data_max!=0)
      Shutdown();
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
void CDictionary::FreeMode(const bool mode)
  {
   if(m_free_mode==mode)
      return;
   m_free_mode=mode;
//--- communicate freemode to CList objects
   for(int i=0;i<m_data_max;i++)
     {
      if(m_data[i]==NULL)
         continue;
      m_data[i].FreeMode(mode);
     }
  }
//+------------------------------------------------------------------+
//| Clearing of dictionary without the release of memory             |
//+------------------------------------------------------------------+
template<typename T>
void CDictionary::Clear(void)
  {
//--- "physical" removal of the object (if necessary and possible)
   if(m_free_mode)
     {
      for(int i=0;i<m_data_max;i++)
        {
         if(CheckPointer(m_data[i])==POINTER_DYNAMIC)
            delete m_data[i];
         m_data[i]=NULL;
        }
     }
   m_data_total=0;
  }
//+------------------------------------------------------------------+
//| Complete cleaning of the dictionary with the release of memory   |
//+------------------------------------------------------------------+
template<typename T>
bool CDictionary::Shutdown(void)
  {
//--- check
   if(m_data_max==0)
      return(true);
//--- clean
   Clear();
   ArrayFree(m_data);
   m_data_max=0;
//--- successful
   return(true);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
bool CDictionary::Add(CDictionaryEntry<T>*entry)
  {
   if(CheckPointer(entry))
     {
      int index=Index(Hash(entry.Key()));
      CList *list=m_data[index];
      if(CheckPointer(list))
        {
         CDictionaryEntry<T>*result=list.Search(entry);
         if(CheckPointer(result))
           {
            //--- reconnect iteration chain 
            ReconnectChain(result);
            list.Delete(list.IndexOf(result));
            m_data_total--;
           }
        }
      else
        {
         list=new CList();
         m_data[index]=list;
        }
      if(list.Add(entry)>=0)
        {
         list.Sort(0);
         m_data_total++;
         AddModelToChain(entry);
         return true;
        }
     }
   return false;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
bool CDictionary::Allocate(void)
  {
   if(m_data_total>m_data_max*m_resize_threshold)
     {
      int new_size= ArrayResize(m_data,m_data_max*m_multiplier);
      if(new_size!=m_data_max*m_multiplier)

         return false;
      m_data_max=new_size;
      return Recalculate();
     }
   return true;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
template<typename KEY>
bool CDictionary::Contains(KEY key)
  {
   string _key=string(key);
   bool res=false;
   T value=NULL;
   int index=Index(Hash(_key));
   CList *list=m_data[index];
   if(CheckPointer(list))
     {
      CDictionaryEntry<T> *model=new CDictionaryEntry<T>(_key,value);
      if(CheckPointer(list.Search(model)))
         res=true;
      delete model;
     }
   return res;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
template<typename KEY>
bool CDictionary::Delete(KEY key)
  {
   string _key=string(key);
   bool res=false;
   T value=NULL;
   int index=Index(Hash(_key));
   CList *list=m_data[index];
   if(CheckPointer(list))
     {
      CDictionaryEntry<T>*model=new CDictionaryEntry<T>(_key,value);
      CDictionaryEntry<T>*result=list.Search(model);
      if(CheckPointer(result))
        {
         //--- reconnect iteration chain
         ReconnectChain(result);
         //--- delete the item
         list.Delete(list.IndexOf(result));
         res=true;
        }
      delete model;
     }
   return res;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
template<typename KEY>
bool CDictionary::Get(KEY key,T &value)
  {
   string _key=string(key);
   bool res=false;
   int index=Index(Hash(_key));
   CList *list=m_data[index];
   if(CheckPointer(list))
     {
      CDictionaryEntry<T>*model=new CDictionaryEntry<T>(_key,value);
      CDictionaryEntry<T>*result=list.Search(model);
      if(CheckPointer(result))
        {
         CDictionaryEntry<T>*model_entry=result;
         value=model_entry.Value();
         res=true;
        }
      delete model;
     }
   return res;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
template<typename KEY>
T CDictionary::Get(KEY key)
  {
   T value;
   Get((string)key,value);
   return value;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
uint CDictionary::Hash(string str)
  {
   uint fnv_prime=16777619;
   uint offset_basis=2166136261;
   uint hash=offset_basis;
   uchar arr[];
   StringToCharArray(str,arr);
   uint len=StringLen(str);
   for(uint n=0; n<len; n++)
     {
      hash^= arr[n];
      hash =(hash*fnv_prime);
     }
   return hash%INT_MAX;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
int CDictionary::Index(uint key)
  {
   return int(key % m_data_max);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
bool CDictionary::Recalculate(void)
  {
   CArrayObj array;
   array.FreeMode(false);
   m_data_total=0;
   for(int i=0;i<m_data_max;i++)
     {
      CList *list=m_data[i];
      if(!CheckPointer(list))
         continue;
      while(list.Total()!=0)
         array.Add(list.DetachCurrent());
     }
   for(int i=0;i<array.Total();i++)
     {
      CDictionaryEntry<T>*entry=array.At(i);
      if(CheckPointer(entry))
        {
         int index=Index(Hash(entry.Key()));
         CList *list=m_data[index];
         if(!CheckPointer(list))
           {
            list=new CList();
            m_data[index]=list;
           }
         if(list.Add(entry)>=0)
           {
            list.Sort(0);
            m_data_total++;
            AddModelToChain(entry);
           }
        }
     }
   return true;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
bool CDictionary::Reset(void)
  {
   if(m_free_mode && m_data_total>0)
      for(int i=0;i<m_data_max;i++)
         if(CheckPointer(m_data[i]))
            delete m_data[i];
   m_data_total=ArrayResize(m_data,m_initial_capacity);
   m_data_max=m_data_total;
   return true;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+

//template<typename T>
//template<typename KEY>
//bool CDictionary::Add(KEY key,T value)
//{
//   return Set(key,value);
//}
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
template<typename KEY>
bool CDictionary::Add(KEY key,T value)
  {
   string _key=string(key);
   if(Allocate())
     {
      int index=Index(Hash(_key));
      CDictionaryEntry<T>*model=new CDictionaryEntry<T>(_key,value);
      CList *list=m_data[index];
      if(CheckPointer(list))
        {
         CDictionaryEntry<T>*result=list.Search(model);
         if(CheckPointer(result))
           {
            //--- reconnect iteration chain 
            ReconnectChain(result);
            if(list.Delete(list.IndexOf(result)))
               m_data_total--;
            else
              {
               delete model;
               return false;
              }
           }
         if(list.Add(model)>=0)
           {
            list.Sort(0);
            m_data_total++;
            AddModelToChain(model);
            return true;
           }
        }
      else
        {
         list=new CList();
         m_data[index]=list;
         if(list.Add(model)>=0)
           {
            list.Sort(0);
            m_data_total++;
            AddModelToChain(model);
           }
         return true;
        }
     }
   return false;
  }
//+------------------------------------------------------------------+
//| Returns current object. Returns NULL if an object was not        |
//| NULL.                                                            |
//+------------------------------------------------------------------+
template<typename T>
CDictionaryEntry<T>*CDictionary::GetCurrentNode(void)
  {
   return m_current_entry;
  }
//+------------------------------------------------------------------+
//| Returns the previous object. After call of method current        |
//| object becomes the previous one. If object not selected, returns |
//| NULL.                                                            |
//+------------------------------------------------------------------+
template<typename T>
CDictionaryEntry<T>*CDictionary:: GetPrevNode(void)
  {
   if(m_current_entry==NULL)
      return NULL;
   if(m_current_entry.m_prev_entry==NULL)
      return NULL;
   CDictionaryEntry<T>*entry=m_current_entry.m_prev_entry;
   m_current_entry=entry;
   return entry;
  }
//+------------------------------------------------------------------+
//| Returns the next object.     After call of method current        |
//| object becomes the next one. If object not selected, returns     |
//| NULL.                                                            |
//+------------------------------------------------------------------+
template<typename T>
CDictionaryEntry<T>*CDictionary::GetNextNode(void)
  {
   if(m_current_entry==NULL)
      return NULL;
   if(m_current_entry.m_next_entry==NULL)
      return NULL;
   CDictionaryEntry<T>*entry=m_current_entry.m_next_entry;
   m_current_entry=entry;
   return entry;
  }
//+------------------------------------------------------------------+
//| Returns first node from the node list. If dictionary has no nodes|
//| returns NULL.                                                    |
//+------------------------------------------------------------------+
template<typename T>
CDictionaryEntry<T>*CDictionary::GetFirstNode(void)
  {
   m_current_entry=m_first_entry;
   return m_first_entry;
  }
//+------------------------------------------------------------------+
//| Returns the last node from the node list. If no nodes in dict,   |
//| it returns NULL.                                                 |
//+------------------------------------------------------------------+
template<typename T>
CDictionaryEntry<T>*CDictionary::GetLastNode(void)
  {
   m_current_entry=m_last_entry;
   return m_last_entry;
  }
//+------------------------------------------------------------------+
//| Reconnect both sides of a deleted item in iteration chain        |
//+------------------------------------------------------------------+
template<typename T>
bool CDictionary::ReconnectChain(CDictionaryEntry<T>*deleted_item)
  {
//--- reconnect chain in both sides of deleted item
   CDictionaryEntry<T> *p_entry = deleted_item.m_prev_entry;
   CDictionaryEntry<T> *n_entry = deleted_item.m_next_entry;
   if(m_first_entry.Compare(deleted_item)==0)
      m_first_entry=n_entry;
   if(m_last_entry.Compare(deleted_item)==0)
      m_last_entry=p_entry;
   if(CheckPointer(p_entry))
      p_entry.m_next_entry=n_entry;
   if(CheckPointer(n_entry))
      n_entry.m_prev_entry=p_entry;
   return true;
  }
//+------------------------------------------------------------------+
//| Add model to end of iteration chain                              |
//+------------------------------------------------------------------+
template<typename T>
bool CDictionary::AddModelToChain(CDictionaryEntry<T>*model)
  {
//--- we add to the end 
   m_current_entry=model;
   if(m_last_entry==NULL)
     {
      m_last_entry=model;
      m_first_entry=model;
      m_current_entry=model;
     }
   else
     {
      model.m_prev_entry=m_last_entry;
      m_last_entry.m_next_entry=model;
      m_last_entry=model;
     }
   return true;
  }
//+------------------------------------------------------------------+
